local savedVars
local lastTime = 0
local defaults = {
   enableGuild = true,
   enableGroup = true,
   enableFriend = true,
   enableAll = true,
   enableLevel = true,
   enableHealth = true,
   enableHealthPercent = true,
   fontFace = "Univers 67",
   fontSize = 20,
   fontOutline = "soft-shadow-thin",
}
local guildCache = {}
local groupCache = {}

-- Maintains list of group mebers
local function OnUnitUpdate(eventCode, unitTag)
   if ZO_Group_IsGroupUnitTag(unitTag) then
      local value = eventCode == EVENT_UNIT_CREATED or nil
      groupCache[GetRawUnitName(unitTag)] = value
   end
end

local function RebuildGuildCache()
   guildCache = {}
   local numGuilds = GetNumGuilds()
   for guildIndex = 1, numGuilds do
      local guildId = GetGuildId(guildIndex)
      local numMembers = GetNumGuildMembers(guildId)
      for memberIndex = 1, numMembers do
         local hasCharacter, characterName = GetGuildMemberCharacterInfo(guildId, memberIndex)
         if hasCharacter then
            guildCache[characterName] = true
         end
      end
   end
end

-- If Reticle is hidden then hide labels
local function OnReticleHidden()
   RETICLE.reticleTexture:SetColor(1,1,1)
   RETICLE.stealthIcon.stealthEyeTexture:SetColor(1,1,1)
   RainbowReticleUnitName:SetHidden(true)
end

-- Returns formated name
local function GetFormatedUnitName(unitTag, name)
   if savedVars.enableLevel == true then
      if IsUnitVeteran(unitTag) then 
         return zo_strformat("<<C:1>> (VR<<2>>)", name, GetUnitVeteranRank(unitTag))
      end
      return zo_strformat("<<C:1>> <<2[//($d)]>>", name, GetUnitLevel(unitTag))
   end

   return zo_strformat(SI_UNIT_NAME, name) 
end

-- local function IsUnitGuildMember(unitTag)
--    local target = GetUnitRawName(unitTag)
--    for i = 1, GetNumGuilds() do
--       for j = 1, GetNumGuildMembers(xguildid) do
--          local hasCharacter, characterName = GetGuildMemberCharacterInfo(i, j)
--          if hasCharacter and characterName == target then
--             return true
--          end
--       end
--    end
--    return false
-- end

-- local function IsUnitGroupMember(unitTag)
--    if IsUnitGrouped('player') and IsUnitGrouped(unitTag) then
--       for i = 1, GetGroupSize() do
--          local groupTag = GetGroupUnitTagByIndex(i)
--          if AreUnitsEqual(unitTag, groupTag) then
--             return true
--          end
--       end
--    end
--    return false
-- end

local function GetUnitNameAndColor(unitTag)
   local color = ZO_ColorDef:New()
   local target = GetRawUnitName(unitTag)   
   local name = ""
   
   if IsUnitPlayer(unitTag) and GetUnitReaction(unitTag) == UNIT_REACTION_PLAYER_ALLY then
      if IsUnitFriend(unitTag) then
         color:SetRGB(.2,.2,1)
         if savedVars.enableFriend then
            name = GetFormatedUnitName(unitTag, target)
         end
         return name, color  
      end
      

      if guildCache[target] then
         color:SetRGB(.2,1,.2) 
         if savedVars.enableGuild then
            name = GetFormatedUnitName(unitTag, target)
         end
         return name, color
      end
         
      if groupCache[target] then
         if savedVars.enableGroup then
            name = GetFormatedUnitName(unitTag, target)
         end
         return name, color
      end
   end
   
   color:SetRGB(GetUnitReactionColor(unitTag))
   if savedVars.enableAll then
      local current, max = GetUnitPower(unitTag, POWERTYPE_HEALTH)
      if max > 1 then
         name = GetFormatedUnitName(unitTag, target)
      end
   end

   return name, color   
end

-- If target has changed check it's a valid target then find out if they are friend of guildy
local function OnTargetHasChanged(eventcode, invname)
   local text, color 

   if DoesUnitExist('reticleover') and #GetUnitName('reticleover') > 0 then
      text, color = GetUnitNameAndColor('reticleover')
      RETICLE.stealthIcon.stealthText:SetColor(color:UnpackRGB())
   else
      text = ""
      color = ZO_ColorDef:New()
      RETICLE.stealthIcon.stealthText:SetColor(ZO_HIGHLIGHT_TEXT:UnpackRGB())
   end   

   RETICLE.stealthIcon.stealthTextTimeline:PlayForward()

   RETICLE.stealthIcon.stealthEyeTexture:SetColor(color:UnpackRGB())
   RETICLE.reticleTexture:SetColor(color:UnpackRGB())

   if #text > 0 then
      RainbowReticleUnitName:SetColor(color:UnpackRGB())
      RainbowReticleUnitName:SetText(text)
      RainbowReticleUnitName:SetHidden(false)
   else
      RainbowReticleUnitName:SetHidden(true)
   end

   if savedVars.enableHealth or savedVars.enableHealthPercent then
      RainbowReticleUnitHealth:SetColor(color:UnpackRGB())
      RainbowReticleUnitHealth:SetText("")
   end
end

local function OnUpdate()
   local now = GetFrameTimeMilliseconds()
   if (now - lastTime) < 100 then return end
   lastTime = now

   if not DoesUnitExist('reticleover') then return end

   local target = GetUnitName('reticleover')

   if #target > 0 then
      local text = ""
      if IsUnitDead('reticleover') then
         text = "DEAD"
      elseif IsUnitInvulnerableGuard('reticleover') then
         text = "Invulnerable"
      else
         local current, max = GetUnitPower('reticleover', POWERTYPE_HEALTH)
         if max > 10 then
            local percent = zo_round(current / max * 100)
            if savedVars.enableHealth then
               text = ("%d/%d"):format(current, max)
            end 
            if savedVars.enableHealthPercent then
               if #text > 0 then
                  text = ("%s (%d%%)"):format(text, percent)
               else
                  text = percent.."%"
               end
            end
         end
      end
      RainbowReticleUnitHealth:SetText(text)
   end
end

local function InitOnUpdate()
   if savedVars.enableHealth or savedVars.enableHealthPercent then
      RainbowReticle:SetHandler("OnUpdate", OnUpdate)
      RainbowReticleUnitHealth:SetHidden(false)
   else
      RainbowReticle:SetHandler("OnUpdate", nil)
      RainbowReticleUnitHealth:SetHidden(true)
   end
end

local function InitGroupCache()
   if savedVars.enableGroup then
      if IsUnitGrouped('player') then
         for i = 1, GetGroupSize() do
            local unitTag = GetGroupUnitTagByIndex(i)
            groupCache[GetRawUnitName(unitTag)] = true
         end
      end
      EVENT_MANAGER:RegisterForEvent("RainbowReticle", EVENT_UNIT_CREATED, OnUnitUpdate)
      EVENT_MANAGER:RegisterForEvent("RainbowReticle", EVENT_UNIT_DESTROYED, OnUnitUpdate)
   else
      groupCache = {}
      EVENT_MANAGER:UnregisterForEvent("RainbowReticle", EVENT_UNIT_CREATED)
      EVENT_MANAGER:UnregisterForEvent("RainbowReticle", EVENT_UNIT_DESTROYED)
   end
end

local function InitGuildCache()
   if savedVars.enableGuild then
      RebuildGuildCache()
      EVENT_MANAGER:RegisterForEvent("RainbowReticle", EVENT_GUILD_SELF_JOINED_GUILD, RebuildGuildCache)
      EVENT_MANAGER:RegisterForEvent("RainbowReticle", EVENT_GUILD_SELF_LEFT_GUILD, RebuildGuildCache)
      EVENT_MANAGER:RegisterForEvent("RainbowReticle", EVENT_GUILD_MEMBER_ADDED, RebuildGuildCache)
      EVENT_MANAGER:RegisterForEvent("RainbowReticle", EVENT_GUILD_MEMBER_REMOVED, RebuildGuildCache)
      EVENT_MANAGER:RegisterForEvent("RainbowReticle", EVENT_GUILD_MEMBER_CHARACTER_UPDATED, RebuildGuildCache)
   else
      guildCache = {}
      EVENT_MANAGER:UnregisterForEvent("RainbowReticle", EVENT_GUILD_SELF_JOINED_GUILD)
      EVENT_MANAGER:UnregisterForEvent("RainbowReticle", EVENT_GUILD_SELF_LEFT_GUILD)
      EVENT_MANAGER:UnregisterForEvent("RainbowReticle", EVENT_GUILD_MEMBER_ADDED)
      EVENT_MANAGER:UnregisterForEvent("RainbowReticle", EVENT_GUILD_MEMBER_REMOVED)
      EVENT_MANAGER:UnregisterForEvent("RainbowReticle", EVENT_GUILD_MEMBER_CHARACTER_UPDATED)
   end
end

local function SetFont()
   local font = ("%s|%d|%s"):format(LibStub("LibMediaProvider-1.0"):Fetch("font", savedVars.fontFace), savedVars.fontSize, savedVars.fontOutline)
   RainbowReticleUnitName:SetFont(font)
   RainbowReticleUnitHealth:SetFont(font)
end

local function CreateSettingsMenu()
   local panelData = {
      type = "panel",
      name = 'Rainbow Reticle',
      displayName = '|cFF0000R|r|cFFFF00a|r|cFF33CCi|r|c00CC00n|r|c0033CCb|r|cCC3300o|r|c663300w|r Reticle',
      author = "Mitazaki & Garkin",
      version = "1.1",
      slashCommand = "/reticle",
      registerForRefresh = true,
      registerForDefaults = true,
   }

   local optionsData = {
      {
         type = "header",
         name = ZO_HIGHLIGHT_TEXT:Colorize("NAMES"),
      },
      {
         type = "checkbox",
         name = 'Enable Guild Names',
         tooltip = 'If enabled player names of guild mates will be displayed',
         getFunc = function() return savedVars.enableGuild end,
         setFunc = function(value)
               savedVars.enableGuild = value
               InitGuildCache()
               RainbowReticleUnitName:SetHidden(not (savedVars.enableGuild or savedVars.enableFriend or defaults.enableGroup or savedVars.enableAll))
            end,
         default = defaults.enableGuild,
      },
      {
         type = "checkbox",
         name = 'Enable Friend Names',
         tooltip = 'If enabled player names of friends will be displayed',
         getFunc = function() return savedVars.enableFriend end,
         setFunc = function(value)
               savedVars.enableFriend = value
               RainbowReticleUnitName:SetHidden(not (savedVars.enableGuild or savedVars.enableFriend or defaults.enableGroup or savedVars.enableAll))
            end,
         default = defaults.enableFriend,
      },
      {
         type = "checkbox",
         name = 'Enable Group Names',
         tooltip = 'Will display player names of group members.',
         getFunc = function() return savedVars.enableGroup end,
         setFunc = function(value)
               savedVars.enableGroup = value
               InitGroupCache()
               RainbowReticleUnitName:SetHidden(not (savedVars.enableGuild or savedVars.enableFriend or defaults.enableGroup or savedVars.enableAll))
            end,
         default = defaults.enableGroup,
      },
      {
         type = "checkbox",
         name = 'Enable Other Names',
         tooltip = 'Will display names for all other targets',
         getFunc = function() return savedVars.enableAll end,
         setFunc = function(value)
               savedVars.enableAll = value
               RainbowReticleUnitName:SetHidden(not (savedVars.enableGuild or savedVars.enableFriend or defaults.enableGroup or savedVars.enableAll))
            end,
         default = defaults.enableAll,
      },
      {
         type = "checkbox",
         name = 'Enable Level Indicator',
         tooltip = 'Will display level indicator for enabled targets',
         getFunc = function() return savedVars.enableLevel end,
         setFunc = function(value) savedVars.enableLevel = value end,
         default = defaults.enableLevel,
      },
      {
         type = "header",
         name = ZO_HIGHLIGHT_TEXT:Colorize("HEALTH INDICATOR"),
      },
      {
         type = "checkbox",
         name = 'Enable Health Values',
         tooltip = 'If enabled targets health min/max numbers will be displayed',
         getFunc = function() return savedVars.enableHealth end,
         setFunc = function(value) savedVars.enableHealth = value; InitOnUpdate() end,
         default = defaults.enableHealth,
      },
      {
         type = "checkbox",
         name = 'Enable Health Percentages',
         tooltip = 'If enabled targets health percentages will be displayed',
         getFunc = function() return savedVars.enableHealthPercent end,
         setFunc = function(value) savedVars.enableHealthPercent = value; InitOnUpdate() end,
         default = defaults.enbaleHealthPercent
      },
      {
         type = "header",
         name = ZO_HIGHLIGHT_TEXT:Colorize("FONT"),
      },
      {
         type    = "dropdown",
         name    = "Font Face",
         choices = LibStub("LibMediaProvider-1.0"):List("font"),
         getFunc = function() return savedVars.fontFace end,
         setFunc = function(value) savedVars.fontFace = value; SetFont() end,
         default = defaults.fontFace,
      },
      {
         type    = "dropdown",
         name    = "Font Outline",
         choices = { "none", "outline", "thin-outline", "thick-outline", "shadow", "soft-shadow-thin", "soft-shadow-thick" },
         getFunc = function() return savedVars.fontOutline end,
         setFunc = function(value) savedVars.fontOutline = value; SetFont() end,
         default = defaults.fontOutline,
      },
      {
         type    = "slider",
         name    = "Font Size",
         min     = 8,
         max     = 72,
         getFunc = function() return savedVars.fontSize end,
         setFunc = function(value) savedVars.fontSize = value; SetFont() end,
         default = defaults.fontSize,
      },
      {
         type = "header",
         name = ZO_HIGHLIGHT_TEXT:Colorize("ABOUT RAINDOW RETICLE"),
      },
      {
         type = "description",
         text = "      |cC5C29EAuthor:|r Mitazaki |c666666(modified by mctaylor & Garkin)|r\n      |cC5C29EVersion:|r " .. panelData.version .. "\n\n      Copyright 2014 All Rights Reserved",
      },
   }

   local LAM = LibStub("LibAddonMenu-2.0")
   LAM:RegisterAddonPanel('rainbowconfigpanel', panelData)
   LAM:RegisterOptionControls('rainbowconfigpanel', optionsData)
end

-- Init RainbowReticle
local function OnAddOnLoaded(eventCode, addOnName)
   if addOnName ~= "RainbowReticle" then return end
   EVENT_MANAGER:UnregisterForEvent("RainbowReticle", EVENT_ADD_ON_LOADED)

   savedVars = ZO_SavedVars:NewAccountWide("RAINBOWRETICLE_DB", 2, nil, defaults)

   SetFont()
   InitOnUpdate()
   InitGroupCache()
   InitGuildCache()
   CreateSettingsMenu()
   
   --reticle
   EVENT_MANAGER:RegisterForEvent("RainbowReticle", EVENT_RETICLE_TARGET_CHANGED, OnTargetHasChanged)
   EVENT_MANAGER:RegisterForEvent("RainbowReticles", EVENT_RETICLE_HIDDEN_UPDATE, OnReticleHidden)
end

EVENT_MANAGER:RegisterForEvent("RainbowReticle", EVENT_ADD_ON_LOADED, OnAddOnLoaded)
